/*
 * Double linked list.
 *
 * This software is distributed under the terms of GNU
 * Public License version 2 or later versions.
 *
 */
#include <stdlib.h>
#ifndef _LIST_H
#define _LIST_H

struct list_item {
	struct list_item	*next;
	struct list_item	*prev;
	void			*data;
};

struct list {
	struct list_item	*head;
	struct list_item	*tail;
};

typedef int (*list_compare_t)(const void *a, const void *b);
typedef int (*list_callback_t)(struct list_item *item, void *opaque);

#define LIST_END(list)		((struct list_item *) list)
#define LIST_INIT(list)		((list)->head = (list)->tail = LIST_END(list));

#define list_foreach(list, iter) \
  for ((iter) = (list)->head; iter != LIST_END(list); (iter) = (iter)->next)

#define list_first(list)	((list)->head)
#define list_last(list)		((list)->tail)

#define list_is_empty(list)	\
  ((list)->tail == (struct list_item *) list && \
   (list)->head == (struct list_item *) (list))

/* Create list head. */
struct list *list_new();

/* Free all elements of the list. */
void list_free(struct list **list)
{
	struct list_item *iter;
	if (!list_is_empty(*list))
	{
		list_foreach(*list, iter)
		{
			free(iter);
		}
	}
}

/* Add element to the head of the list. */
void list_add(struct list *list, struct list_item *item)
{
        if (list_is_empty(list)) 
	{
        list_first(list) = item;
	list_last(list) = item;
	}
}

/* Add element to the tail of the list. */
void list_add_tail(struct list *list, struct list_item *item)
{
	 if (!list_is_empty(list)) 
	{
		struct list_item *add_item = (struct list_item *)malloc(sizeof(struct list_item));
		list_last(list)->next = add_item;
		list_last(list)->next = item;
		(list)->tail = list_last(list)->next;		
	}
}

/* Add data to to the tail of the list. */
void list_add_data(struct list *list, void *data)
{
	struct list_item *iter;
	if (!list_is_empty(list))
	{
		(list)->tail->data = data;
	}

}

/* Delete data from the list. */
void list_del_data(struct list *list, void *data)
{
        struct list_item *iter;
	if (!list_is_empty(list))
	{
		list_foreach(list, iter)
       		{
                	if ((iter)->data == data)                        	
			free((iter)->data);
		}
	}
}

/* Unlink entry from the list. */
struct list_item *list_unlink(struct list *list, struct list_item *item)
{
        struct list_item *iter;
	if (!list_is_empty(list))
	{
		list_foreach(list, iter)
		{
                	if (iter == item)
                	{
				((iter)->prev)->next = (iter)->next;
                        	free(iter);
                	} 
		}
	}
}

/* Get pointer to n-th element of the list. */
struct list_item *list_nth(struct list *list, unsigned int nth)
{
	struct list_item *iter;
	unsigned int i = 0;
	if (!list_is_empty(list))
	{
		list_foreach(list, iter)
		{
			if (i == nth)
	       		break;
	       		i++;
		}
	}
}

/* Returns number of elements in the list. */
size_t list_count(struct list *list)
{
	size_t count = 0;
	struct list_item *iter;
	if (!list_is_empty(list))
	{
		list_foreach(list, iter)
		{
			count++;
		}
	}
	return (count);
}

/* Sort list using compare function. */
void list_sort(struct list *list, list_compare_t compare)
{
	if (!list_is_empty(list))
	qsort(list, list_count(list), sizeof(struct list_item), compare);
}



/* Traverse list forward and backward. */
void list_traverse(struct list *list, list_callback_t callback, void *opaque)
{
	struct list_item *iter;
	int rc;
	if (!list_is_empty(list))
	{
	list_foreach(list, iter)
	{
		rc = (*callback)(iter, opaque);
		if (rc != 0)
                        break;
	}
	}
}

void list_traverse_backward(struct list *list, list_callback_t callback, void *opaque)
{
	struct list_item *iter;
	int rc;
	for ((iter) = (list)->tail; iter != list_first(list); (iter) = (iter)->prev)
	{
		rc = (*callback)(iter, opaque);
		if (rc != 0)
                        break;
	}
}

#endif
